import logging
from time import time

from bubuku.broker import BrokerManager
from bubuku.communicate import sleep_and_operate
from bubuku.env_provider import EnvProvider
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.controller')


class Change(object):
    def get_name(self) -> str:
        raise NotImplementedError('Not implemented yet')

    def can_run(self, current_actions) -> bool:
        raise NotImplementedError('Not implemented yet')

    def run(self, current_actions) -> bool:
        raise NotImplementedError('Not implemented')

    def can_run_at_exit(self) -> bool:
        return False

    def on_remove(self):
        pass


class Check(object):
    def __init__(self, check_interval_s=5):
        self.check_interval_s = check_interval_s
        self.__last_check_timestamp_s = 0

    def check_if_time(self) -> Change:
        if self.time_till_check() <= 0:
            self.__last_check_timestamp_s = time()
            _LOG.info('Executing check {}'.format(self))
            return self.check()
        return None

    def time_till_check(self):
        return self.__last_check_timestamp_s + self.check_interval_s - time()

    def check(self) -> Change:
        raise NotImplementedError('Not implemented')


def _exclude_self(provider_id, name, running_actions):
    return [k for k, v in running_actions.items() if k != name or v != provider_id]


class Controller(object):
    def __init__(self, broker_manager: BrokerManager, zk: BukuExhibitor, env_provider: EnvProvider):
        self.broker_manager = broker_manager
        self.zk = zk
        self.env_provider = env_provider
        self.checks = []
        self.changes = {}  # Holds mapping from change name to array of pending changes
        self.running = True
        self.provider_id = None  # provider id must not be requested on initialization

    def enumerate_changes(self):
        with self.zk.lock(self.provider_id):
            running_changes = self.zk.get_running_changes()

        result = []
        for name, change_list in self.changes.items():
            running = running_changes.get(name) == self.provider_id
            first = True
            for change in change_list:
                result.append({
                    'type': name,
                    'description': str(change),
                    'running': bool(first and running)
                })
                first = False
        return result

    def cancel_changes(self, name):
        result = len(self.changes.get(name, {}))
        if result:
            with self.zk.lock(self.provider_id):
                self.zk.unregister_change(name)

            del self.changes[name]
        return result

    def add_check(self, check):
        _LOG.info('Adding check {}'.format(str(check)))
        self.checks.append(check)

    def _register_running_changes(self) -> dict:
        if not self.changes:
            return {}  # Do not take lock if there are no changes to register
        _LOG.debug('Taking lock for processing')
        with self.zk.lock(self.provider_id):
            _LOG.debug('Lock is taken')
            # Get list of current running changes
            running_changes = self.zk.get_running_changes()
            if running_changes:
                _LOG.info("Running changes: {}".format(running_changes))
            # Register changes to run
            for name, change_list in self.changes.items():
                # Only first change is able to run
                first_change = change_list[0]
                if first_change.can_run(_exclude_self(self.provider_id, name, running_changes)):
                    if name not in running_changes:
                        self.zk.register_change(name, self.provider_id)
                        running_changes[name] = self.provider_id
                else:
                    _LOG.info('Change {} is waiting for others: {}'.format(name, running_changes))
            return running_changes

    def _run_changes(self, running_changes: dict) -> list:
        changes_to_remove = []
        for name, change_list in self.changes.items():
            if name in running_changes and running_changes[name] == self.provider_id:
                change = change_list[0]
                _LOG.info('Executing action {} step'.format(change))
                if self.running or change.can_run_at_exit():
                    try:
                        if not change.run(_exclude_self(self.provider_id, change.get_name(), running_changes)):
                            _LOG.info('Action {} completed'.format(change))
                            changes_to_remove.append(change.get_name())
                        else:
                            _LOG.info('Action {} will be executed on next loop step'.format(change))
                    except Exception as e:
                        _LOG.error('Failed to execute change {} because of exception, removing'.format(change),
                                   exc_info=e)
                        changes_to_remove.append(change.get_name())
                else:
                    _LOG.info(
                        'Action {} can not be run while stopping, forcing to stop it'.format(change))
                    changes_to_remove.append(change.get_name())
        return changes_to_remove

    def _release_changes_lock(self, changes_to_remove):
        if changes_to_remove:
            for change_name in changes_to_remove:
                removed_change = self.changes[change_name][0]
                del self.changes[change_name][0]
                if not self.changes[change_name]:
                    del self.changes[change_name]
                removed_change.on_remove()
            with self.zk.lock():
                for name in changes_to_remove:
                    self.zk.unregister_change(name)

    def loop(self, change_on_init=None):
        self.provider_id = self.env_provider.get_id()
        if change_on_init:
            self._add_change_to_queue(change_on_init)
        while self.running or self.changes:
            self.make_step()

            if self.changes:
                timeout = 0.5
            else:
                timeout = min([check.time_till_check() for check in self.checks])
            sleep_and_operate(self, timeout)

    def make_step(self):
        # register running changes
        running_changes = self._register_running_changes()
        # apply changes without holding lock
        changes_to_remove = self._run_changes(running_changes)
        # remove processed actions
        self._release_changes_lock(changes_to_remove)
        if self.running:
            for check in self.checks:
                self._add_change_to_queue(check.check_if_time())

    def _add_change_to_queue(self, change):
        if not change:
            return
        _LOG.info('Adding change {} to pending changes'.format(change.get_name()))
        if change.get_name() not in self.changes:
            self.changes[change.get_name()] = []
        self.changes[change.get_name()].append(change)

    def stop(self, change: Change):
        _LOG.info('Stopping controller with additional change: {}'.format(change.get_name() if change else None))
        # clear all pending changes
        self._add_change_to_queue(change)
        self.running = False



import logging

from instance_control import node
from instance_control import piu
from instance_control.aws import AWSResources
from instance_control.command import Command

_LOG = logging.getLogger('bubuku.cluster.command.terminate')


class TerminateCommand(Command):
    def __init__(self, cluster_config_path: str, ip: str, user: str, odd: str):
        super().__init__(cluster_config_path)
        self.ip = ip
        self.user = user
        self.odd = odd

    def execute(self):
        aws_ = AWSResources(region=self.cluster_config['region'])
        instance = node.get_instance_by_ip(aws_.ec2_resource, self.cluster_config['cluster_name'], self.ip)
        piu.stop_taupage(self.ip, self.user, self.odd, self.cluster_config['region'])
        node.terminate(aws_, self.cluster_config['cluster_name'], instance)



import json
import logging
import time

from botocore.exceptions import ClientError

from instance_control.aws import AWSResources

_LOG = logging.getLogger('bubuku.cluster.aws.iam')


def create_or_get_instance_profile(aws_: AWSResources, cluster_config: dict):
    profile_name = 'profile-{}'.format(cluster_config['cluster_name'])
    role_name = 'role-{}'.format(cluster_config['cluster_name'])
    policy_datavolume = 'policy-{}-datavolume'.format(cluster_config['cluster_name'])
    policy_metadata = 'policy-{}-metadata'.format(cluster_config['cluster_name'])
    policy_zmon = 'policy-{}-zmon'.format(cluster_config['cluster_name'])

    try:
        profile = aws_.iam_client.get_instance_profile(InstanceProfileName=profile_name)
        _LOG.info("IAM profile %s exists, using it", profile_name)
        return profile['InstanceProfile']
    except ClientError:
        _LOG.info("IAM profile %s does not exists, creating ...", profile_name)
        pass

    profile = aws_.iam_client.create_instance_profile(InstanceProfileName=profile_name)

    _LOG.info("Creating iam role %s", role_name)
    aws_.iam_client.create_role(RoleName=role_name, AssumeRolePolicyDocument="""{
        "Version": "2012-10-17",
        "Statement": [{
             "Action": "sts:AssumeRole",
             "Effect": "Allow",
             "Principal": {
                 "Service": "ec2.amazonaws.com"
             }
        }]
    }""")

    policy_datavolume_document = """{
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": [
                    "ec2:DescribeTags",
                    "ec2:DeleteTags",
                    "ec2:DescribeVolumes",
                    "ec2:AttachVolume"
                ],
                "Resource": "*"
            }
        ]
    }"""
    policy_metadata_document = """{
        "Version": "2012-10-17",
        "Statement": [
            {
                "Action": "ec2:Describe*",
                "Resource": "*",
                "Effect": "Allow"
            },
            {
                "Action": "elasticloadbalancing:Describe*",
                "Resource": "*",
                "Effect": "Allow"
            }
        ]
    }"""
    policy_zmon_document = """{
        "Version": "2012-10-17",
        "Statement": [
            {
                "Action": "cloudwatch:PutMetricData",
                "Resource": "*",
                "Effect": "Allow"
            }
        ]
    }"""

    _LOG.info("Creating IAM policy %s", policy_datavolume)
    aws_.iam_client.put_role_policy(RoleName=role_name,
                                    PolicyName=policy_datavolume,
                                    PolicyDocument=policy_datavolume_document)
    _LOG.info("Creating IAM policy %s", policy_metadata)
    aws_.iam_client.put_role_policy(RoleName=role_name,
                                    PolicyName=policy_metadata,
                                    PolicyDocument=policy_metadata_document)
    _LOG.info("Creating IAM policy %s", policy_zmon)
    aws_.iam_client.put_role_policy(RoleName=role_name,
                                    PolicyName=policy_zmon,
                                    PolicyDocument=policy_zmon_document)

    if "kms_key_id" in cluster_config:
        policy_kms_document = json.dumps({
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Action": "kms:Decrypt",
                    "Effect": "Allow",
                    "Resource": [cluster_config["kms_key_id"]]
                }
            ]
        })
        _LOG.info("Creating IAM policy $s", policy_kms_document)
        aws_.iam_client.put_role_policy(RoleName=role_name,
                                        PolicyName='policy-{}-kms'.format(cluster_config['cluster_name']),
                                        PolicyDocument=policy_kms_document)

    aws_.iam_client.add_role_to_instance_profile(InstanceProfileName=profile_name, RoleName=role_name)

    _LOG.info("IAM profile %s is created", profile_name)

    #
    # FIXME: using an instance profile right after creating one
    # can result in 'not found' error, because of eventual
    # consistency.  For now fix with a sleep, should rather
    # examine exception and retry after some delay.
    #
    _LOG.info('Waiting 30 secs after IAM profile creation to be sure it is available')
    time.sleep(30)

    return profile['InstanceProfile']



import logging
from typing import List

from bubuku.features.rebalance import BaseRebalanceChange
from bubuku.zookeeper import BukuExhibitor

FAKE_ZONE = 'fake_zone'

FAKE_BROKER_ID = 'fake_id'


class Partition(object):
    def __init__(self, topic: str, partition: str, brokers: List['Broker']):
        self.topic = topic
        self.partition = partition
        self.brokers = brokers

    def get_key(self):
        return self.topic, self.partition

    def change_assignment(self, from_: 'Broker', to: 'Broker'):
        if not to.can_accept_partition(self):
            return False
        for idx in range(0, len(self.brokers)):
            if self.brokers[idx] != from_:
                continue
            from_.remove_partition(idx == 0, self)
            to.add_partition(idx == 0, self)
            self.brokers[idx] = to
            return True
        raise Exception('Inconsistency because of poor data structure model2')

    def optimize_leaders(self, only_solve_overload=True):
        leader = self.brokers[0]
        if only_solve_overload and not leader.is_leaders_overloaded():
            return False
        smallest_load = None
        smallest_load_idx = None
        for idx in range(1, len(self.brokers)):
            load_factor = self.brokers[idx].get_load_factor(True, 1)
            if smallest_load is None or smallest_load > load_factor:
                smallest_load = load_factor
                smallest_load_idx = idx
        if smallest_load_idx is None:
            return False
        less_loaded = self.brokers[smallest_load_idx]

        if less_loaded.is_leaders_overloaded(1):
            return False

        if leader.leader_count - less_loaded.leader_count < 2:
            return False

        leader.remove_partition(True, self)
        if not leader.add_partition(False, self):
            raise Exception('Can not add partition {}'.format(self.get_key()))

        less_loaded.remove_partition(False, self)
        if not less_loaded.add_partition(True, self):
            raise Exception('Can not add partition {}'.format(self.get_key()))

        self.brokers[0] = less_loaded
        self.brokers[smallest_load_idx] = leader
        return True


class Broker(object):
    def __init__(self, id_: str, zone: str):
        self.id_ = id_
        self.zone = zone
        self.replica_count = 0
        self.leader_count = 0
        self.expected_replica_count = (0, 0)
        self.expected_leader_count = (0, 0)
        self.partitions = {}
        self.healthy = True

    def get_id(self):
        return self.id_

    def __str__(self):
        return str(self.id_)

    def set_partitions_estimation(self, replica_counts, leader_count):
        self.expected_replica_count = replica_counts
        self.expected_leader_count = leader_count

    def get_load_factor(self, for_leader, add=0):
        if for_leader:
            return float(self.leader_count + add) / self.expected_leader_count[1]
        else:
            return float(self.replica_count + add) / self.expected_replica_count[1]

    def is_leaders_overloaded(self, add=0):
        return (self.leader_count + add) > self.expected_leader_count[1]

    def is_replicas_overloaded(self, add=0):
        return (self.replica_count + add) > self.expected_replica_count[1]

    def list_partition_keys(self):
        result = []
        for pkey, leaderArray in self.partitions.items():
            for is_leader in leaderArray:
                result.append((pkey, is_leader))
        return result

    def is_leader_for(self, pkey):
        return any(self.partitions.get(pkey, []))

    def copy_partition_keys(self):
        return [k for k in self.partitions.keys()]

    def iterate(self, leader_only: bool):
        for pkey, leader_array in self.partitions.items():
            if leader_only:
                if any(leader_array):
                    yield pkey
            else:
                yield pkey

    def add_partition(self, leader: bool, partition: Partition):
        key = partition.get_key()
        if key in self.partitions:
            return False
        self._add_partition_int(key, leader)
        return True

    def remove_partition(self, leader: bool, partition: Partition):
        key = partition.get_key()
        if key not in self.partitions:
            raise Exception('Trying to detach partition that was not attached to broker')
        data = self.partitions[key]
        for idx in range(0, len(data)):
            if bool(leader) == bool(data[idx]):
                del data[idx]
                if leader:
                    self.leader_count -= 1
                self.replica_count -= 1
                if not data:
                    del self.partitions[key]
                return
        raise Exception("Well, something bad happen. Partition is removed, but it is not a leader or whatever.")

    def _add_partition_int(self, key, leader):
        if key not in self.partitions:
            self.partitions[key] = []
        self.partitions[key].append(leader)
        if leader:
            self.leader_count += 1
        self.replica_count += 1

    def can_accept_partition(self, partition: Partition):
        return partition.get_key() not in self.partitions


class ZoneChecker(object):
    def __init__(self, zones: List[str]):
        self.zones = zones
        self.expected_distributions = {}

    def is_allowed_move(self, from_: Broker, to: Broker, partition: Partition):
        current_zone_distribution = {z: 0 for z in self.zones}
        for b in partition.brokers:
            if b.zone not in self.zones:
                continue  # This actually removes fake one
            current_zone_distribution[b.zone] = current_zone_distribution.get(b.zone, 0) + 1
        if from_.zone in current_zone_distribution:
            current_zone_distribution[from_.zone] -= 1
        if 0 == current_zone_distribution.get(to.zone, 0):
            return True
        # Allow move only to zone with minimal amount of partitions
        return current_zone_distribution[to.zone] == min(current_zone_distribution.values())

    def _get_expected_distribution(self, count: int) -> List[int]:
        if count not in self.expected_distributions:
            min_per_zone = count // len(self.zones)
            delta = count - min_per_zone * len(self.zones)
            result = [min_per_zone + 1 if i < delta else min_per_zone for i in range(0, len(self.zones))]
            self.expected_distributions[count] = result
        return self.expected_distributions[count].copy()

    def get_improper_brokers(self, brokers) -> List[Broker]:
        expected = sorted(self._get_expected_distribution(len(brokers)))
        current_distribution = {}
        for b in brokers:
            current_distribution[b.zone] = current_distribution.get(b.zone, 0) + 1
        broken_zones = {}
        for zone in [k for k in current_distribution.keys() if k in self.zones]:
            value = current_distribution[zone]
            delta = value - expected[-1]
            if delta >= 0:
                expected = expected[:-1]
            if delta > 0:
                broken_zones[zone] = delta
        result = []
        for zone, count_to_remove in broken_zones.items():
            result += [x for x in brokers if x.zone == zone][:count_to_remove]
        return result


class FakeBroker(Broker):
    def __init__(self):
        super(FakeBroker, self).__init__(FAKE_BROKER_ID, FAKE_ZONE)
        self.healthy = False

    def add_partition(self, leader: bool, partition: Partition):
        self._add_partition_int(partition.get_key(), leader)
        return True


def transfer_partition(partition, from_: Broker, target_brokers_sorted: List[Broker],
                       zone_checker: ZoneChecker, polite=False):
    for target in target_brokers_sorted:
        if not zone_checker.is_allowed_move(from_, target, partition):
            continue
        if polite and target.is_replicas_overloaded(1):
            continue
        if partition.change_assignment(from_, target):
            return True
    return False


def get_division_result_min_max(nom, denom):
    if denom == 0:
        return 0, 0
    return nom // denom, ((nom - 1) // denom) + 1


_LOG = logging.getLogger('simple_rebalance')


class SimpleRebalanceChange(BaseRebalanceChange):
    _STATE_INIT = 'init'
    _STATE_DISTRIBUTE_AMONG = 'distribute'
    _STATE_EMPTY_FAKE = 'empty_fake'
    _STATE_OPTIMIZE_REPLICAS = 'optimize_replicas'
    _STATE_OPTIMIZE_LEADERS = 'optimize_leaders'
    _STATE_BALANCE = 'balance'

    def __init__(self, zk: BukuExhibitor, broker_ids: list, empty_brokers: list, exclude_topics: list,
                 parallelism: int):
        self.state = self._STATE_INIT
        self.zk = zk
        self.fake = FakeBroker()
        self.active_brokers = {}
        self.partitions = {}
        self.rebalance_queue = {}
        self.parallelism = parallelism
        self.exclude_topics = exclude_topics if exclude_topics else []
        self.empty_brokers = [str(e) for e in empty_brokers] if empty_brokers else []
        self.initial_broker_ids = sorted([str(broker_id) for broker_id in broker_ids])
        self.zone_checker = None

    def register_partition_change(self, partition: Partition):
        self.rebalance_queue[(partition.topic, partition.partition)] = partition

    def load_data_from_zk(self):
        existing_brokers = []
        for broker_id, broker_rack in self.zk.get_broker_racks().items():
            str_broker_id = str(broker_id)
            existing_brokers.append(str_broker_id)
            if str_broker_id in self.empty_brokers:
                continue
            self.active_brokers[str_broker_id] = Broker(str_broker_id, str(broker_rack))

        for topic, partition, broker_ids in self.zk.load_partition_assignment():
            if topic in self.exclude_topics:
                continue
            if not any([str(id_) in existing_brokers for id_ in broker_ids]):
                _LOG.warning('Partition {} {} is lost, cause no broker have it'.format(topic, partition))
                continue
            p = Partition(topic, str(partition), [self.active_brokers.get(str(id_), self.fake) for id_ in broker_ids])
            self.partitions[p.get_key()] = p

    def createInitialDistribution(self):
        total_partitions = 0
        total_leaders = 0
        az_counts = {}
        for broker in self.active_brokers.values():
            az_counts[broker.zone] = az_counts.get(broker.zone, 0) + 1
        self.zone_checker = ZoneChecker([zone for zone in az_counts.keys()])

        for partition in self.partitions.values():
            replication_factor = len(partition.brokers)

            improper_brokers = self.zone_checker.get_improper_brokers(partition.brokers)
            for broker_idx in range(0, len(partition.brokers)):
                added = False
                broker = partition.brokers[broker_idx]
                if broker.get_id() in self.active_brokers and broker not in improper_brokers:
                    if broker.add_partition(broker_idx == 0, partition):
                        added = True
                if not added:
                    partition.brokers[broker_idx] = self.fake
                    self.fake.add_partition(broker_idx == 0, partition)
                    self.register_partition_change(partition)

            total_partitions += replication_factor
            total_leaders += 1

        leaders_min, leaders_max = get_division_result_min_max(total_leaders, len(self.active_brokers))

        per_zone_partitions_min, per_zone_partitions_max = get_division_result_min_max(total_partitions, len(az_counts))

        for zone, count in az_counts.items():
            per_broker_partitions_min, _ = get_division_result_min_max(per_zone_partitions_min, count)
            _, per_broker_partitions_max = get_division_result_min_max(per_zone_partitions_max, count)
            for broker in self.active_brokers.values():
                if broker.zone == zone:
                    broker.set_partitions_estimation(
                        (per_broker_partitions_min, per_broker_partitions_max),
                        (leaders_min, leaders_max),
                    )

    def empty_fake(self):
        leaders_load = sorted(self.active_brokers.values(), key=lambda x: x.get_load_factor(True))
        replica_load = sorted(self.active_brokers.values(), key=lambda x: x.get_load_factor(False))

        for pkey, is_leader in self.fake.list_partition_keys():
            load = leaders_load if is_leader else replica_load
            partition = self.partitions[pkey]
            if not transfer_partition(partition, self.fake, load, self.zone_checker, polite=False):
                raise Exception('Nowhere to transfer partition {}'.format(pkey))
            self.register_partition_change(partition)
            if is_leader:
                leaders_load = sorted(leaders_load, key=lambda x: x.get_load_factor(True))
            else:
                replica_load = sorted(replica_load, key=lambda x: x.get_load_factor(False))

    def optimize_replicas(self):
        current_load = sorted(self.active_brokers.values(), key=lambda x: x.get_load_factor(False))
        moved = True
        while moved:
            to_empty = current_load[-1]
            if not to_empty.is_replicas_overloaded():
                return
            moved = False
            for pkey in to_empty.copy_partition_keys():
                partition = self.partitions[pkey]
                if not transfer_partition(partition, to_empty, current_load, self.zone_checker, polite=True):
                    continue
                moved = True
                self.register_partition_change(partition)
                if not to_empty.is_replicas_overloaded():
                    break
            if moved:
                current_load = sorted(current_load, key=lambda x: x.get_load_factor(False))

    def optimize_leaders(self):
        # First step, try to balance leaders for partitions that we are already moving
        has_modifications = True
        while has_modifications:
            has_modifications = False
            for p in self.rebalance_queue.values():
                if p.optimize_leaders():
                    has_modifications = True
            if not has_modifications:
                for p in self.rebalance_queue.values():
                    if p.optimize_leaders(only_solve_overload=False):
                        has_modifications = True
            if not has_modifications:
                for p in self.partitions.values():
                    if not p.optimize_leaders():
                        continue
                    self.register_partition_change(p)
                    has_modifications = True
            if not has_modifications:
                for p in self.partitions.values():
                    if not p.optimize_leaders(only_solve_overload=False):
                        continue
                    self.register_partition_change(p)
                    has_modifications = True

    def perform_rebalance(self):
        to_rebalance = [k for k in self.rebalance_queue.keys()]
        if len(to_rebalance) > self.parallelism:
            to_rebalance = to_rebalance[:self.parallelism]
        to_rebalance = [self.rebalance_queue.pop(k) for k in to_rebalance]
        if not to_rebalance:
            return True
        to_rebalance_data = [
            (p.topic, int(p.partition), [b.id_ for b in p.brokers])
            for p in to_rebalance
        ]
        if not self.zk.reallocate_partitions(to_rebalance_data):
            for partition in to_rebalance:
                self.register_partition_change(partition)
        return False

    def run(self, current_actions) -> bool:
        # Stop rebalance if someone is restarting
        if self.should_be_paused(current_actions):
            _LOG.warning("Rebalance paused, because other blocking events running: {}".format(current_actions))
            return True
        if self.zk.is_rebalancing():
            return True

        new_broker_ids = sorted([str(id_) for id_ in self.zk.get_broker_ids()])
        if new_broker_ids != self.initial_broker_ids:
            _LOG.warning("Rebalance stopped because of broker list change from {} to {}".format(
                self.initial_broker_ids, new_broker_ids))
            return False

        if self.state == SimpleRebalanceChange._STATE_INIT:
            # Load data from zk
            self.load_data_from_zk()
            if not self.active_brokers:
                _LOG.info("no active brokers, can not do rebalance")
                return False
            self.state = SimpleRebalanceChange._STATE_DISTRIBUTE_AMONG
        elif self.state == SimpleRebalanceChange._STATE_DISTRIBUTE_AMONG:
            # Sort and distribute to entities
            self.createInitialDistribution()
            self.state = SimpleRebalanceChange._STATE_EMPTY_FAKE
        elif self.state == SimpleRebalanceChange._STATE_EMPTY_FAKE:
            # Move data around from non-existent brokers
            self.empty_fake()
            self.state = SimpleRebalanceChange._STATE_OPTIMIZE_REPLICAS
        elif self.state == SimpleRebalanceChange._STATE_OPTIMIZE_REPLICAS:
            # Now try to evenly distribute partitions/leaders among brokers
            self.optimize_replicas()
            self.state = SimpleRebalanceChange._STATE_OPTIMIZE_LEADERS
        elif self.state == SimpleRebalanceChange._STATE_OPTIMIZE_LEADERS:
            # Now try to evenly distribute partitions/leaders among brokers
            self.optimize_leaders()
            self.state = SimpleRebalanceChange._STATE_BALANCE
        elif self.state == SimpleRebalanceChange._STATE_BALANCE:
            rebalance_finished = self.perform_rebalance()
            return not rebalance_finished
        else:
            _LOG.warning("Stopping rebalance, as state {} is not supported".format(self.state))
            return False
        return True

    def __str__(self):
        return 'SimpleRebalance state={}, queue_size={}, parallelism={}'.format(
            self.state, len(self.rebalance_queue), self.parallelism)



import logging

import subprocess

_LOG = logging.getLogger('bubuku.cluster.piu')


def stop_taupage(ip: str, user: str, odd: str, region: str):
    _LOG.info('Stopping taupage container on %s', ip)
    piu = ["piu", ip, "--region", region, "-O", odd, "\"detaching ebs, terminate and launch instance\""]
    _call(piu)

    stop = ["ssh", "-tA", user + '@' + odd, "ssh", "-o", "StrictHostKeyChecking=no", user + '@' + ip,
            "'docker stop -t 300 taupageapp'"]
    _call(stop)
    _LOG.info('Taupage container on %s is successfully stopped', ip)


def _call(cmd):
    proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    if proc.returncode != 0 and proc.returncode != 64:
        raise Exception(proc.returncode, stderr.decode('utf-8'))
    return stdout



import logging
from bubuku.broker import BrokerManager
from bubuku.controller import Check
from bubuku.features.rebalance.change import OptimizedRebalanceChange
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.features.rebalance')


class RebalanceOnStartCheck(Check):
    def __init__(self, zk: BukuExhibitor, broker: BrokerManager):
        super().__init__()
        self.zk = zk
        self.broker = broker
        self.executed = False

    def check(self):
        if self.executed:
            return None
        if not self.broker.is_running_and_registered():
            return None
        _LOG.info("Rebalance on start, triggering rebalance")
        self.executed = True
        return OptimizedRebalanceChange(self.zk, self.zk.get_broker_ids(), [], [])

    def __str__(self):
        return 'RebalanceOnStartCheck (executed={})'.format(self.executed)


class RebalanceOnBrokerListCheck(Check):
    def __init__(self, zk: BukuExhibitor, broker: BrokerManager):
        super().__init__()
        self.zk = zk
        self.broker = broker
        self.old_broker_list = []

    def check(self):
        if not self.broker.is_running_and_registered():
            return None
        new_list = self.zk.get_broker_ids()
        if not new_list == self.old_broker_list:
            _LOG.info('Broker list changed from {} to {}, triggering rebalance'.format(self.old_broker_list, new_list))
            self.old_broker_list = new_list
            return OptimizedRebalanceChange(self.zk, new_list, [], [])
        return None

    def __str__(self):
        return 'RebalanceOnBrokerListChange, cached list: {}'.format(self.old_broker_list)



import json
import logging
import threading
from http.server import BaseHTTPRequestHandler, HTTPServer

from bubuku.utils import get_opt_broker_id, prepare_configs
from bubuku.zookeeper import load_exhibitor_proxy


class ApiConfig():
    ENDPOINT = '/api/broker/'
    PORT = 8888

    @staticmethod
    def get_url(ip: str, path: str):
        return 'http://{}:{}{}{}'.format(ip, ApiConfig.PORT, ApiConfig.ENDPOINT, path)


_LOG = logging.getLogger('bubuku.api')


class _Handler(BaseHTTPRequestHandler):
    def do_POST(self):
        if self.path in '{}{}'.format(ApiConfig.ENDPOINT, 'stop'):
            config, env_provider = prepare_configs()
            with load_exhibitor_proxy(env_provider.get_address_provider(), config.zk_prefix) as zk:
                try:
                    broker_id = get_opt_broker_id(None, config, zk, env_provider)
                    from bubuku.features.remote_exec import RemoteCommandExecutorCheck
                    RemoteCommandExecutorCheck.register_stop(zk, broker_id)
                    self._send_response()
                except Exception as e:
                    self._send_response({'state': 'stopped', 'details': str(e)})

    def do_GET(self):
        if self.path in '{}{}'.format(ApiConfig.ENDPOINT, 'state'):
            config, env_provider = prepare_configs()
            with load_exhibitor_proxy(env_provider.get_address_provider(), config.zk_prefix) as zk:
                try:
                    broker_id = get_opt_broker_id(None, config, zk, env_provider)
                    if zk.is_broker_registered(broker_id):
                        if any([c not in zk.get_running_changes() for c in ['start', 'restart', 'stop']]):
                            self._send_response({'state': 'restarting'})
                            return
                        self._send_response({'state': 'running'})
                    else:
                        self._send_response({'state': 'stopped'})
                except Exception as e:
                    self._send_response({'state': 'stopped', 'details': str(e)})

    def _send_response(self, json_=None, status_code=200):
        self.send_response(status_code)
        self.send_header('Content-type', 'application/json')
        self.end_headers()
        self.wfile.write(json.dumps(json_).encode('utf-8'))


def start_api_server() -> threading.Thread:
    def _thread_func():
        server = HTTPServer(('', ApiConfig.PORT), _Handler)
        server.serve_forever()
        server.socket.close()

    t = threading.Thread(target=_thread_func, daemon=True)
    _LOG.info('Starting API server on port {}'.format(ApiConfig.PORT))
    t.start()
    return t



import logging
import subprocess

from bubuku.config import Config, KafkaProperties, load_config
from bubuku.env_provider import EnvProvider
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.utils')


class CmdHelper(object):
    def get_disk_stats(self) -> (int, int):
        """
        Returns total disk stats,
        :return: used_kb, free_kb
        """
        disks = self.cmd_run("df -k | tail -n +2 |  awk '{ print $3, $4 }'").split("\n")
        total_used = total_free = 0
        for disk in disks:
            parts = disk.split(" ")
            if len(parts) == 2:
                used, free = tuple(parts)
                total_used += int(used)
                total_free += int(free)
        return total_used, total_free

    def cmd_run(self, cmd):
        output = subprocess.check_output(cmd, shell=True)
        return output.decode("utf-8")


def get_opt_broker_id(broker_id: str, config: Config, zk: BukuExhibitor, env_provider: EnvProvider) -> str:
    if not broker_id:
        kafka_properties = KafkaProperties(config.kafka_settings_template, '/tmp/tmp.props'.format(config.kafka_dir))
        broker_id_manager = env_provider.create_broker_id_manager(zk, kafka_properties)
        broker_id = broker_id_manager.get_broker_id()
        _LOG.info('Will use broker_id {}'.format(broker_id))
    running_brokers = zk.get_broker_ids()
    if broker_id not in running_brokers:
        raise Exception('Broker id {} is not registered ({})'.format(broker_id, running_brokers))
    return broker_id


def prepare_configs():
    config = load_config()
    _LOG.info('Using config: {}'.format(config))
    env_provider = EnvProvider.create_env_provider(config)
    return config, env_provider



import json
import logging
import threading
from functools import partial
from http.server import BaseHTTPRequestHandler, HTTPServer

from bubuku.communicate import execute_on_controller_thread
from bubuku.controller import Controller
from bubuku.utils import CmdHelper

_CONTROLLER_TIMEOUT = 5

_API_CONTROLLER = '/api/controller/'

_LOG = logging.getLogger('bubuku.health')


def load_controller_queue(controller: Controller):
    return controller.enumerate_changes()


def delete_from_controller_queue(name: str, controller: Controller):
    return {
        'count': controller.cancel_changes(name)
    }


class _Handler(BaseHTTPRequestHandler):
    cmd_helper = None

    def do_GET(self):
        if self.path in ('/api/disk_stats', '/api/disk_stats/'):
            used_kb, free_kb = self.cmd_helper.get_disk_stats()
            self._send_response({'free_kb': free_kb, 'used_kb': used_kb})
        elif self.path.startswith(_API_CONTROLLER):
            self.wrap_controller_execution(lambda: self._run_controller_action(self.path[len(_API_CONTROLLER):]))
        else:
            self._send_response({'status': 'OK'})

    def wrap_controller_execution(self, call):
        try:
            call()
        except TimeoutError as e:
            _LOG.error('Failed to rum action because of timeouts', exc_info=e)
            self._send_response({'code': 'timeout', 'message': 'Timeout occurred'}, 500)

    def do_DELETE(self):
        if not self.path.startswith(_API_CONTROLLER):
            return self._send_response({'message': 'Path {} is not supported'.format(self.path)}, 404)
        action = self.path[len(_API_CONTROLLER):].split('/')
        if action[0] == 'queue':
            if len(action) < 2:
                return self._send_response({'message': 'No second argument provided!'}, 400)
            self.wrap_controller_execution(
                lambda: self._send_response(execute_on_controller_thread(
                    partial(delete_from_controller_queue, action[1]), _CONTROLLER_TIMEOUT), 200))
        else:
            return self._send_response({'message': 'Action {} is not supported'.format(action[0])}, 404)

    def _run_controller_action(self, action):
        if action.split('/')[0] == 'queue':
            return self._send_response(execute_on_controller_thread(load_controller_queue, _CONTROLLER_TIMEOUT), 200)
        else:
            return self._send_response({'message': 'Action {} is not supported'.format(action)}, 404)

    def _send_response(self, json_, status_code=200):
        self.send_response(status_code)
        self.send_header('Content-type', 'application/json')
        self.end_headers()
        self.wfile.write(json.dumps(json_).encode('utf-8'))


def start_server(port, cmd_helper: CmdHelper) -> threading.Thread:
    def _thread_func():
        _Handler.cmd_helper = cmd_helper
        server = HTTPServer(('', port), _Handler)
        server.serve_forever()
        server.socket.close()

    t = threading.Thread(target=_thread_func, daemon=True)
    _LOG.info('Starting health server on port {}'.format(port))
    t.start()
    return t



import logging

from bubuku.broker import BrokerManager
from bubuku.controller import Check
from bubuku.utils import CmdHelper
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.features.data_size_stats')


class GenerateDataSizeStatistics(Check):
    def __init__(self, zk: BukuExhibitor, broker: BrokerManager, cmd_helper: CmdHelper, kafka_log_dirs: list):
        super().__init__(check_interval_s=600)
        self.zk = zk
        self.broker = broker
        self.cmd_helper = cmd_helper
        self.kafka_log_dirs = kafka_log_dirs

    def check(self):
        if self.broker.is_running_and_registered():
            _LOG.info("Generating data size statistics")
            try:
                self.__generate_stats()
                _LOG.info("Data size statistics successfully written to zk")
            except Exception:
                _LOG.warn("Error occurred when collecting size statistics", exc_info=True)
        return None

    def __str__(self):
        return 'GenerateDataSizeStatistics'

    def __generate_stats(self):
        topics_stats = self.__get_topics_stats()
        used_kb, free_kb = self.cmd_helper.get_disk_stats()
        stats = {"disk": {'used_kb': used_kb, 'free_kb': free_kb}, "topics": topics_stats}
        self.zk.update_disk_stats(self.broker.id_manager.get_broker_id(), stats)

    def __get_topics_stats(self):
        topics_stats = {}
        for log_dir in self.kafka_log_dirs:
            _LOG.info("Processing log dir: {}".format(log_dir))
            topic_dirs = self.cmd_helper.cmd_run("du -k -d 1 {}".format(log_dir)).split("\n")
            for topic_dir in topic_dirs:
                dir_stats = self.__parse_dir_stats(topic_dir, log_dir)
                if dir_stats:
                    topic, partition, size_kb = dir_stats
                    if topic not in topics_stats:
                        topics_stats[topic] = {}
                    topics_stats[topic][partition] = int(size_kb)
        return topics_stats

    @staticmethod
    def __parse_dir_stats(topic_dir, log_dir):
        """
        Parses topic-partition size stats from "du" tool single line output
        :param topic_dir: the string to be parsed; example: "45983\t/tmp/kafka-logs/my-kafka-topic-0"
        :param log_dir: the kafka log directory name itself
        :return: tuple (topic, partition, size) or None if the topic_dir has incorrect format
        """
        dir_data = topic_dir.split("\t")
        if len(dir_data) == 2 and dir_data[1] != log_dir:
            size_kb, dir_name = tuple(dir_data)
            tp_name = dir_name.split("/")[-1]
            tp_parts = tp_name.rsplit("-", 1)
            if len(tp_parts) == 2:
                topic, partition = tuple(tp_parts)
                return topic, partition, size_kb
        return None



#!/usr/bin/env python3
import logging
import os
import re

from bubuku.config import KafkaProperties
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.id_generator')


class BrokerIdGenerator(object):
    def __init__(self, zk: BukuExhibitor, kafka_properties: KafkaProperties):
        super().__init__()
        self.zk = zk
        self.kafka_properties = kafka_properties
        self.broker_id = None

    def get_broker_id(self):
        if self.broker_id:
            return self.broker_id

        meta_path = '{}/meta.properties'.format(self.kafka_properties.get_property('log.dirs'))
        while not os.path.isfile(meta_path):
            return None
        with open(meta_path) as f:
            lines = f.readlines()
            for line in lines:
                match = re.search('broker\.id=(\d+)', line)
                if match:
                    self.broker_id = match.group(1)
                    return self.broker_id
        return None

    def is_registered(self):
        broker_id = self.get_broker_id()
        if broker_id:
            return self.zk.is_broker_registered(broker_id)
        return False

